# coding: utf-8
"""
This module contains objects for postprocessing A2F calculations (phonon lifetimes in metals
and Eliashberg function).

Warning:
    Work in progress, DO NOT USE THIS CODE.
"""
from __future__ import annotations

import itertools
import numpy as np
import pandas as pd
import pymatgen.core.units as units
import abipy.core.abinit_units as abu

from collections import OrderedDict
from scipy.integrate import cumtrapz, simps
#from typing import Any
from monty.string import marquee, list_strings
from monty.functools import lazy_property
from abipy.core.structure import Structure
from abipy.core.mixins import AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter
from abipy.core.kpoints import Kpath
from abipy.tools.plotting import (add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_axlims, set_visible,
                                  rotate_ticklabels)
from abipy.tools import duck
from abipy.tools.typing import Figure
from abipy.electrons.ebands import ElectronBands, ElectronDos, RobotWithEbands
from abipy.dfpt.phonons import PhononBands, PhononDos, RobotWithPhbands
from abipy.abio.robots import Robot
from abipy.eph.common import BaseEphReader


_LATEX_LABELS = {
    "lambda_iso": r"$\lambda_{iso}$",
    "omega_log": r"$\omega_{log}$",
    "a2f": r"$\alpha^2F(\omega)$",
    "lambda": r"$\lambda(\omega)$",
}


class A2f:
    """
    Eliashberg function a2F(w). Energies are in eV.
    """
    # Markers used for up/down bands.
    marker_spin = {0: "^", 1: "v"}

    def __init__(self, mesh, values_spin, values_spin_nu, ngqpt, meta):
        """
        Args:
            mesh: Energy mesh in eV
            values(nomega,0:natom3,nsppol)
            vals(w,1:natom,1:nsppol): a2f(w) decomposed per phonon branch and spin
            vals(w,0,1:nsppol): a2f(w) summed over phonons modes, decomposed in spin
            ngqpt: Q-mesh used to compute A2f.
            meta: Dictionary with metavariables.

        TODO:
            1. possibility of computing a2f directly from data on file?
        """
        self.mesh = mesh
        self.ngqpt = ngqpt
        self.meta = meta

        # Spin dependent and total a2F(w)
        values_spin = np.atleast_2d(values_spin)
        values_spin_nu = np.atleast_3d(values_spin_nu)
        self.nsppol = len(values_spin)
        self.nmodes = values_spin_nu.shape[1]
        assert self.nmodes % 3 == 0
        self.natom = self.nmodes // 3

        if self.nsppol == 2:
            self.values = values_spin[0] + values_spin[1]
            self.values_nu = values_spin_nu[0] + values_spin_nu[1]
        elif self.nsppol == 1:
            self.values = values_spin[0]
            self.values_nu = values_spin_nu[0]
        else:
            raise ValueError("Invalid nsppol: %s" % self.nsppol)

        self.values_spin = values_spin
        self.values_spin_nu = values_spin_nu
        #self.lambdaw ?

    @lazy_property
    def iw0(self) -> int:
        """
        Index of the first point in the mesh whose value is >= 0
        Integrals are performed with wmesh[iw0 + 1, :] i.e. unstable modes are neglected.
        """
        for i, x in enumerate(self.mesh):
            if x >= 0.0: return i
        else:
            raise ValueError("Cannot find zero in energy mesh")

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, title: str = None, verbose: int = 0) -> str:
        """
        String representation with verbosity level ``verbose`` and an optional ``title``.
        """
        lines = []; app = lines.append

        app("Eliashberg Function" if not title else str(title))
        # TODO: Add ElectronDos
        app("Isotropic lambda: %.2f, omega_log: %.3f (eV), %.3f (K)" % (
            self.lambda_iso, self.omega_log, self.omega_log * abu.eV_to_K))
        app("Q-mesh: %s" % str(self.ngqpt))
        app("Mesh from %.4f to %.4f (eV) with %d points" % (
            self.mesh[0], self.mesh[-1], len(self.mesh)))

        if verbose:
            for mustar in (0.1, 0.12, 0.2):
                app("\tFor mustar %s: McMillan Tc: %s [K]" % (mustar, self.get_mcmillan_tc(mustar)))

        if verbose > 1:
            # $\int dw [a2F(w)/w] w^n$
            for n in [0, 4]:
                app("Moment %s: %s" % (n, self.get_moment(n)))

            app("Meta: %s" % str(self.meta))

        return "\n".join(lines)

    @lazy_property
    def lambda_iso(self) -> float:
        """Isotropic lambda."""
        return self.get_moment(n=0)

    @lazy_property
    def omega_log(self) -> float:
        r"""
        Logarithmic moment of alpha^2F: exp((2/\lambda) \int dw a2F(w) ln(w)/w)
        """
        iw = self.iw0 + 1
        wmesh, a2fw = self.mesh[iw:], self.values[iw:]

        fw = a2fw / wmesh * np.log(wmesh)
        integral = simps(fw, x=wmesh)

        return np.exp(1.0 / self.lambda_iso * integral)

    def get_moment(self, n: int, spin=None, cumulative: bool = False) -> float:
        r"""
        Computes the moment of a2F(w) i.e. $\int dw [a2F(w)/w] w^n$
        From Allen PRL 59 1460 (See also Grimvall, Eq 6.72 page 175)
        """
        wmesh = self.mesh[self.iw0+1:]
        if spin is None:
            a2fw = self.values[self.iw0+1:]
        else:
            a2fw = self.values_spin[spin][self.iw0+1:]

        # Primitive is given on the same mesh as self.
        ff = a2fw * (wmesh ** (n - 1))
        vals = np.zeros(self.mesh.shape)
        vals[self.iw0+1:] = cumtrapz(ff, x=wmesh, initial=0.0)

        return vals if cumulative else vals[-1].copy()

    def get_moment_nu(self, n: int, nu: int, spin=None, cumulative: bool = False) -> float:
        r"""
        Computes the moment of a2F(w) i.e. $\int dw [a2F(w)/w] w^n$
        From Allen PRL 59 1460 (See also Grimvall, Eq 6.72 page 175)
        """
        wmesh = self.mesh[self.iw0+1:]
        if spin is None:
            a2fw = self.values_nu[nu][self.iw0+1:]
        else:
            a2fw = self.values_spin_nu[spin][nu][self.iw0+1:]

        # Primitive is given on the same mesh as self.
        ff = a2fw * (wmesh ** (n - 1))
        vals = np.zeros(self.mesh.shape)
        vals[self.iw0+1:] = cumtrapz(ff, x=wmesh, initial=0.0)

        return vals if cumulative else vals[-1].copy()

    def get_mcmillan_tc(self, mustar: float) -> float:
        """
        Computes the critical temperature Tc in K computed with the McMillan equation
        and the input mustar.
        """
        tc = (self.omega_log / 1.2) * \
            np.exp(-1.04 * (1.0 + self.lambda_iso) / (self.lambda_iso - mustar * (1.0 + 0.62 * self.lambda_iso)))

        return tc * abu.eV_to_K

    def get_mustar_from_tc(self, tc: float) -> float:
        """
        Return the value of mustar that gives the critical temperature ``tc`` in Kelving
        in the McMillan equation.
        """
        l = self.lambda_iso
        num = l + (1.04 * (1 + l) / np.log(1.2 * abu.kb_eVK * tc / self.omega_log))

        return num / (1 + 0.62 * l)

    @add_fig_kwargs
    def plot(self, what="a2f", units="eV", exchange_xy=False, ax=None,
             xlims=None, ylims=None, label=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot a2F(w) or lambda(w) depending on the value of `what`.

        Args:
            what: a2f for a2F(w), lambda for lambda(w)
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz"). Case-insensitive.
            exchange_xy: True to exchange x-y axes.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            ylims: Limits for y-axis. See xlims for API.
            label: True to add legend label to each curve.
            fontsize: Legend and title fontsize
            kwargs: linestyle, color, linewidth passed to ax.plot.

        Returns: |matplotlib-Figure|
        """""
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        wfactor = abu.phfactor_ev2units(units)
        ylabel = _LATEX_LABELS[what]

        style = dict(
            linestyle=kwargs.pop("linestyle", "-"),
            color=kwargs.pop("color", "k"),
            linewidth=kwargs.pop("linewidth", 1),
        )

        # Plot a2f(w)
        if what == "a2f":
            xx, yy = self.mesh * wfactor, self.values
            if exchange_xy: xx, yy = yy, xx
            ax.plot(xx, yy, label=label, **style)

            if self.nsppol == 2:
                # Plot spin resolved a2f(w).
                for spin in range(self.nsppol):
                    xx, yy = self.mesh * wfactor, self.values_spin[spin]
                    if exchange_xy: xx, yy = yy, xx
                    ax.plot(xx, yy, marker=self.marker_spin[spin], **style)

        # Plot lambda(w)
        elif what == "lambda":
            lambda_w = self.get_moment(n=0, cumulative=True)
            xx, yy = self.mesh * wfactor, lambda_w
            if exchange_xy: xx, yy = yy, xx
            ax.plot(xx, yy, label=label, **style)

        else:
            raise ValueError("Invalid value for what: `%s`" % str(what))

        xlabel = abu.wlabel_from_units(units)
        if exchange_xy: xlabel, ylabel = ylabel, xlabel

        ax.set_xlabel(xlabel)
        ax.set_ylabel(ylabel)
        ax.grid(True)
        set_axlims(ax, xlims, "x")
        set_axlims(ax, ylims, "y")
        if label: ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot_with_lambda(self, units="eV", ax=None, xlims=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot a2F(w) and lambda(w) on the same figure.

        Args:
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz"). Case-insensitive.
            ax: |matplotlib-Axes| or None if a new figure should be created.
            xlims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            fontsize: Legend and title fontsize

        Returns: |matplotlib-Figure|
        """""
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        for i, what in enumerate(["a2f", "lambda"]):
            this_ax = ax if i == 0 else ax.twinx()
            self.plot(what=what, ax=this_ax, units=units, fontsize=fontsize, xlims=xlims, show=False, **kwargs)
            if i > 0:
                this_ax.yaxis.set_label_position("right")
                this_ax.grid(False)

        return fig

    @add_fig_kwargs
    def plot_nuterms(self, units="eV", ax_mat=None, with_lambda=True, fontsize=8,
                     xlims=None, ylims=None, label=None, **kwargs) -> Figure:
        """
        Plot a2F(w), lambda(w) and optionally the individual contributions due to the phonon branches.

        Args:
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz").
                Case-insensitive.
            ax_mat: Matrix of axis of shape [natom, 3]. None if a new figure should be created.
            fontsize: Legend and title fontsize.
            xlims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            ylims: Limits for y-axis. See xlims for API.
            label: True to add legend label to each curve.

        Returns: |matplotlib-Figure|
        """""
        # Get ax_mat and fig.
        nrows, ncols = self.natom, 3
        ax_mat, fig, plt = get_axarray_fig_plt(ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=True, squeeze=False)
        ax_mat = np.reshape(ax_mat, (self.natom, 3))

        wfactor = abu.phfactor_ev2units(units)
        wvals = self.mesh * wfactor

        if with_lambda:
            lax_nu = [ax.twinx() for ax in ax_mat.flat]
            # Share axis after creation. Based on
            # https://stackoverflow.com/questions/42973223/how-share-x-axis-of-two-subplots-after-they-are-created
            lax_nu[0].get_shared_x_axes().join(*lax_nu)
            lax_nu[0].get_shared_y_axes().join(*lax_nu)
            for i, ax in enumerate(lax_nu):
                if i == 2: continue
                ax.set_yticklabels([])
                #ax.set_xticklabels([])

        # TODO Better handling of styles
        a2f_style = dict(
            linestyle=kwargs.pop("linestyle", "-"),
            color=kwargs.pop("color", "k"),
            linewidth=kwargs.pop("linewidth", 1),
        )
        lambda_style = a2f_style.copy()
        lambda_style["color"] = "red"


        for idir, iatom in itertools.product(range(3), range(self.natom)):
            nu = idir + 3 * iatom
            ax = ax_mat[iatom, idir]
            ax.grid(True)
            ax.set_title(r"$\nu = %d$" % nu, fontsize=fontsize)
            if idir == 0:
                ax.set_ylabel(r"$\alpha^2F(\omega)$")
            else:
                pass
                # Turn off tick labels
                #ax.set_yticklabels([])
                #ax.set_yticks([])

            if iatom == self.natom - 1:
                ax.set_xlabel(abu.wlabel_from_units(units))
            #set_axlims(ax, xlims, "x")
            #set_axlims(ax, ylims, "y")

            # Plot total a2f(w)
            ax.plot(wvals, self.values_nu[nu], **a2f_style)

            # Plot lambda(w)
            if with_lambda:
                lambdaw_nu = self.get_moment_nu(n=0, nu=nu, cumulative=True)
                lax = lax_nu[nu]
                lax.plot(wvals, lambdaw_nu, **lambda_style)
                if idir == 2:
                    lax.set_ylabel(r"$\lambda_{\nu}(\omega)$", color=lambda_style["color"])

            #if self.nsppol == 2:
            #   # Plot spin resolved a2f(w)
            #   for spin in range(self.nsppol):
            #       ax.plot(self.mesh, self.values_spin_nu[spin, nu],
            #               marker=self.marker_spin[spin], **a2f_style)

        return fig

    @add_fig_kwargs
    def plot_a2(self, phdos, atol=1e-12, **kwargs) -> Figure:
        """
        Grid with 3 plots showing: a2F(w), F(w), a2F(w). Requires phonon DOS.

        Args:
            phdos: |PhononDos|
            atol: F(w) is replaced by atol in a2F(w) / F(w) ratio where :math:`|F(w)|` < atol

        Returns: |matplotlib-Figure|
        """
        phdos = PhononDos.as_phdos(phdos)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=3, ncols=1,
                                                sharex=True, sharey=False, squeeze=True)
        ax_list = ax_list.ravel()

        # Spline phdos onto a2f mesh and compute a2F(w) / F(w)
        f = phdos.spline(self.mesh)
        f = self.values / np.where(np.abs(f) > atol, f, atol)
        ax = ax_list[0]
        ax.plot(self.mesh, f, color="k", linestyle="-")
        ax.grid(True)
        ax.set_ylabel(r"$\alpha^2(\omega)$ [1/eV]")

        # Plot F(w). TODO: This should not be called plot_dos_idos!
        ax = ax_list[1]
        phdos.plot_dos_idos(ax=ax, what="d", color="k", linestyle="-")
        ax.grid(True)
        ax.set_ylabel(r"$F(\omega)$ [states/eV]")

        # Plot a2f
        self.plot(ax=ax_list[2], color="k", linestyle="-", linewidths=2, show=False)

        return fig

    @add_fig_kwargs
    def plot_tc_vs_mustar(self, start=0.1, stop=0.3, num=50, ax=None, **kwargs) -> Figure:
        """
        Plot Tc(mustar)

        Args:
            start: The starting value of the sequence.
            stop: The end value of the sequence
            num (int): optional. Number of samples to generate. Default is 50. Must be non-negative.
            ax: |matplotlib-Axes| or None if a new figure should be created.

        Returns: |matplotlib-Figure|
        """
        # TODO start and stop to avoid singularity in Mc Tc
        mustar_values = np.linspace(start, stop, num=num)
        tc_vals = [self.get_mcmillan_tc(mustar) for mustar in mustar_values]

        ax, fig, plt = get_ax_fig_plt(ax=ax)
        ax.plot(mustar_values, tc_vals, **kwargs)
        ax.set_yscale("log")
        ax.grid(True)
        ax.set_xlabel(r"$\mu^*$")
        ax.set_ylabel(r"$T_c$ [K]")

        return fig


class A2Ftr:
    """
    Transport Eliashberg function a2F(w). Energies are in eV.
    """
    # Markers used for up/down bands (collinear spin)
    marker_spin = {0: "^", 1: "v"}

    def __init__(self, mesh, vals_in, vals_out):
        """
        Args:
            mesh: Energy mesh in eV
            vals_in(nomega,3,3,0:natom3,nsppol):
                Eliashberg transport functions for in and out scattering
            vals_in(w,3,3,1:natom3,1:nsppol): a2f_tr(w) decomposed per phonon branch and spin
            vals_in(w,3,3,0,1:nsppol): a2f_tr(w) summed over phonons modes, decomposed in spin
        """
        self.mesh = mesh

    @lazy_property
    def iw0(self) -> int:
        """
        Index of the first point in the mesh whose value is >= 0
        Integrals are performed with wmesh[iw0 + 1, :] i.e. unstable modes are neglected.
        """
        for i, x in enumerate(self.mesh):
            if x >= 0.0: return i
        else:
            raise ValueError("Cannot find zero in energy mesh")


class A2fFile(AbinitNcFile, Has_Structure, Has_ElectronBands, NotebookWriter):
    """
    This file contains the phonon linewidths, EliashbergFunction, the |PhononBands|,
    the |ElectronBands| and |ElectronDos| on the k-mesh.
    Provides methods to analyze and plot results.

    Usage example:

    .. code-block:: python

        with A2fFile("out_A2F.nc") as ncfile:
            print(ncfile)
            ncfile.ebands.plot()
            ncfile.phbands.plot()

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: A2fFile
    """
    @classmethod
    def from_file(cls, filepath: str) -> A2fFile:
        """Initialize the object from a netcdf_ file."""
        return cls(filepath)

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = A2fReader(filepath)

    def __str__(self) -> str:
        """String representation."""
        return self.to_string()

    def to_string(self, verbose=0) -> str:
        """String representation."""
        lines = []; app = lines.append

        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")
        app(self.structure.to_string(verbose=verbose, title="Structure"))
        app("")
        app(self.ebands.to_string(with_structure=False, verbose=verbose, title="Electronic Bands"))
        app("")
        app(self.phbands.to_string(with_structure=False, verbose=verbose, title="Phonon Bands"))
        app("")
        # E-PH section
        app(marquee("E-PH calculation", mark="="))
        app("K-mesh for electrons:")
        app(self.ebands.kpoints.ksampling.to_string(verbose=verbose))
        if verbose:
            app("Has transport a2Ftr(w): %s" % self.has_a2ftr)
        app("")

        def print_a2f(a2f, title):
            app(title)
            app("Isotropic lambda: %.2f, omega_log: %.3f (eV), %.3f (K)" % (
                a2f.lambda_iso, a2f.omega_log, a2f.omega_log * abu.eV_to_K))
            for mu in (0.1, 0.16):
                tc = a2f.get_mcmillan_tc(mu)
                app(f"Tc[mu={mu}]: {tc} (K)")

        a2f = self.a2f_qcoarse
        print_a2f(a2f, "a2f(w) on the %s q-mesh (ddb_ngqpt|eph_ngqpt)" % str(a2f.ngqpt))
        #app(self.a2f_qcoarse.to_string(title=title, verbose=verbose))
        app("")
        a2f = self.a2f_qintp
        print_a2f(a2f, "a2f(w) Fourier interpolated on the %s q-mesh (ph_ngqpt)" % str(a2f.ngqpt))
        #app("Isotropic lambda: %.2f, omega_log: %.3f (eV), %.3f (K)" % (
        #    a2f.lambda_iso, a2f.omega_log, a2f.omega_log * abu.eV_to_K))
        #app(self.a2f_qintp.to_string(title=title, verbose=verbose))

        return "\n".join(lines)

    @lazy_property
    def ebands(self) -> ElectronBands:
        """|ElectronBands| object."""
        return self.reader.read_ebands()

    @lazy_property
    def edos(self) -> ElectronDos:
        """|ElectronDos| object with e-DOS computed by Abinit."""
        return self.reader.read_edos()

    @property
    def structure(self) -> Structure:
        """|Structure| object."""
        return self.ebands.structure

    @property
    def phbands(self):
        """
        |PhononBands| object with frequencies along the q-path.
        Contains (interpolated) linewidths.
        """
        return self.reader.read_phbands_qpath()

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        od = self.get_ebands_params()
        # Add EPH parameters.
        od.update(self.reader.common_eph_params)

        return od

    @lazy_property
    def a2f_qcoarse(self):
        """
        :class:`A2f` with the Eliashberg function a2F(w) computed on the (coarse) ab-initio q-mesh.
        """
        return self.reader.read_a2f(qsamp="qcoarse")

    @lazy_property
    def a2f_qintp(self):
        """
        :class:`A2f` with the Eliashberg function a2F(w) computed on the dense q-mesh by Fourier interpolation.
        """
        return self.reader.read_a2f(qsamp="qintp")

    def get_a2f_qsamp(self, qsamp):
        """Return the :class:`A2f` object associated to q-sampling ``qsamp``."""
        if qsamp == "qcoarse": return self.a2f_qcoarse
        if qsamp == "qintp": return self.a2f_qintp
        raise ValueError("Invalid value for qsamp `%s`" % str(qsamp))

    @lazy_property
    def has_a2ftr(self) -> bool:
        """True if the netcdf file contains transport data."""
        return "a2ftr_qcoarse" in self.reader.rootgrp.variables

    @lazy_property
    def a2ftr_qcoarse(self):
        """
        :class:`A2ftr` with the Eliashberg transport spectral function a2F_tr(w, x, x')
        computed on the (coarse) ab-initio q-mesh
        """
        if not self.has_a2ftr: return None
        return self.reader.read_a2ftr(qsamp="qcoarse")

    @lazy_property
    def a2ftr_qintp(self):
        """
        :class:`A2ftr` with the Eliashberg transport spectral function a2F_tr(w, x, x')
        computed on the dense q-mesh by Fourier interpolation.
        """
        if not self.has_a2ftr: return None
        return self.reader.read_a2ftr(qsamp="qintp")

    def get_a2ftr_qsamp(self, qsamp):
        """Return the :class:`A2ftr` object associated to q-sampling ``qsamp``."""
        if qsamp == "qcoarse": return self.a2ftr_qcoarse
        if qsamp == "qintp": return self.a2ftr_qintp
        raise ValueError("Invalid value for qsamp `%s`" % str(qsamp))

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    def get_panel(self, **kwargs):
        """
        Build panel with widgets to interact with the A2fFile either in a notebook or in panel app.
        """
        from abipy.panels.a2f import A2fFilePanel
        return A2fFilePanel(ncfile=self).get_panel(**kwargs)

    #def interpolate(self, ddb, lpratio=5, vertices_names=None, line_density=20, filter_params=None, verbose=0):
    #    """
    #    Interpolate the phonon linewidths on a k-path and, optionally, on a k-mesh.

    #    Args:
    #        lpratio: Ratio between the number of star functions and the number of ab-initio k-points.
    #            The default should be OK in many systems, larger values may be required for accurate derivatives.
    #        vertices_names: Used to specify the k-path for the interpolated QP band structure
    #            when ``ks_ebands_kpath`` is None.
    #            It's a list of tuple, each tuple is of the form (kfrac_coords, kname) where
    #            kfrac_coords are the reduced coordinates of the k-point and kname is a string with the name of
    #            the k-point. Each point represents a vertex of the k-path. ``line_density`` defines
    #            the density of the sampling. If None, the k-path is automatically generated according
    #            to the point group of the system.
    #        line_density: Number of points in the smallest segment of the k-path. Used with ``vertices_names``.
    #        filter_params: TO BE DESCRIBED
    #        verbose: Verbosity level

    #    Returns:
    #    """
    #    # Get symmetries from abinit spacegroup (read from file).
    #    abispg = self.structure.abi_spacegroup
    #    fm_symrel = [s for (s, afm) in zip(abispg.symrel, abispg.symafm) if afm == 1]

    #    phbst_file, phdos_file = ddb.anaget_phbst_and_phdos_files(nqsmall=0, ndivsm=10, asr=2, chneut=1, dipdip=1,
    #        dos_method="tetra", lo_to_splitting="automatic", ngqpt=None, qptbounds=None, anaddb_kwargs=None, verbose=0,
    #        mpi_procs=1, workdir=None, manager=None)

    #    phbands = phbst_file.phbands
    #    phbst_file.close()

    #    # Read qibz and ab-initio linewidths from file.
    #    qcoords_ibz = self.reader.read_value("qibz")
    #    data_ibz = self.reader.read_value("phgamma_qibz") * units.Ha_to_eV
    #    import matplotlib.pyplot as plt
    #    plt.plot(data_ibz[0])
    #    plt.show()

    #    # Build interpolator.
    #    from abipy.core.skw import SkwInterpolator
    #    cell = (self.structure.lattice.matrix, self.structure.frac_coords, self.structure.atomic_numbers)

    #    has_timrev = True
    #    fermie, nelect = 0.0, 3 * len(self.structure)
    #    skw = SkwInterpolator(lpratio, qcoords_ibz, data_ibz, fermie, nelect,
    #                          cell, fm_symrel, has_timrev,
    #                          filter_params=filter_params, verbose=verbose)

    #    # Interpolate and set linewidths.
    #    qfrac_coords = [q.frac_coords for q in phbands.qpoints]
    #    phbands.linewidths = skw.interp_kpts(qfrac_coords).eigens

    #    return phbands

    @add_fig_kwargs
    def plot_eph_strength(self, what_list=("phbands", "gamma", "lambda"), ax_list=None,
                          ylims=None, label=None, fontsize=8, **kwargs) -> None:
        """
        Plot phonon bands with EPH coupling strength lambda(q, nu) and lambda(q, nu)
        These values have been Fourier interpolated by Abinit.

        Args:
            what_list: ``phfreqs`` for phonons, `lambda`` for the eph coupling strength,
                ``gamma`` for phonon linewidths.
            ax_list: List of |matplotlib-Axes| (same length as what_list)
                or None if a new figure should be created.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            label: String used to label the plot in the legend.
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        what_list = list_strings(what_list)
        nrows, ncols = len(what_list), 1
        ax_list, fig, plt = get_axarray_fig_plt(ax_list, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = np.array(ax_list).ravel()
        units = "eV"

        for i, (ax, what) in enumerate(zip(ax_list, what_list)):
            # Decorate the axis (e.g add ticks and labels).
            self.phbands.decorate_ax(ax, units="")

            if what == "phbands":
                # Plot phonon bands
                self.phbands.plot(ax=ax, units=units, show=False)
            else:
                # Add eph coupling.
                if what == "lambda":
                    yvals = self.reader.read_phlambda_qpath()
                    ylabel = r"$\lambda(q,\nu)$"
                elif what == "gamma":
                    yvals = self.reader.read_phgamma_qpath()
                    ylabel = r"$\gamma(q,\nu)$ (eV)"
                else:
                    raise ValueError("Invalid value for what: `%s`" % str(what))

                style = dict(
                    linestyle=kwargs.pop("linestyle", "-"),
                    color=kwargs.pop("color", "k"),
                    linewidth=kwargs.pop("linewidth", 1),
                )

                xvals = np.arange(len(self.phbands.qpoints))
                for nu in self.phbands.branches:
                    ax.plot(xvals, yvals[:, nu],
                            label=label if (nu == 0 and label) else None,
                            **style)

                ax.set_ylabel(ylabel)

        set_axlims(ax, ylims, "y")
        if label: ax.legend(loc="best", shadow=True, fontsize=fontsize)

        return fig

    @add_fig_kwargs
    def plot(self, what="gamma", units="eV", scale=None, alpha=0.6, ylims=None, 
             ax=None, colormap="jet", **kwargs) -> Figure:
        """
        Plot phonon bands with gamma(q, nu) or lambda(q, nu) depending on the value of `what`.

        Args:
            what: ``lambda`` for eph coupling strength, ``gamma`` for phonon linewidths.
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz").
                Case-insensitive.
            scale: float used to scale the marker size.
            alpha: The alpha blending value for the markers between 0 (transparent) and 1 (opaque)
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used
            ax: |matplotlib-Axes| or None if a new figure should be created.
            colormap: matplotlib color map.

        Returns: |matplotlib-Figure|
        """
        ax, fig, plt = get_ax_fig_plt(ax=ax)
        cmap = plt.get_cmap(colormap)

        # Plot phonon bands.
        self.phbands.plot(ax=ax, units=units, show=False)

        # Add eph coupling.
        xvals = np.arange(len(self.phbands.qpoints))
        wvals = self.phbands.phfreqs * abu.phfactor_ev2units(units)

        # Sum contributions over nsppol (if spin-polarized)
        # TODO units
        gammas = self.reader.read_phgamma_qpath()
        lambdas = self.reader.read_phlambda_qpath()
        if what == "lambda":
            scale = 500 if scale is None else float(scale)
            sqn = scale * np.abs(lambdas)
            cqn = gammas
        elif what == "gamma":
            scale = 10 ** 6 if scale is None else float(scale)
            sqn = scale * np.abs(gammas)
            cqn = lambdas
        else:
            raise ValueError("Invalid what: `%s`" % str(what))

        vmin, vmax = cqn.min(), cqn.max()

        sc = ax.scatter(np.tile(xvals, len(self.phbands.branches)),
                        wvals.T, # [q, nu] --> [nu, q]
                        s=sqn.T,
                        c=cqn.T,
                        vmin=vmin, vmax=vmax,
                        cmap=cmap,
                        marker="o",
                        alpha=alpha,
                        #label=term if ib == 0 else None
        )

        # Make a color bar
        #plt.colorbar(sc, ax=ax, orientation="horizontal", pad=0.2)
        set_axlims(ax, ylims, "y")

        return fig

    @add_fig_kwargs
    def plot_a2f_interpol(self, units="eV", ylims=None, fontsize=8, **kwargs) -> Figure:
        """
        Compare ab-initio a2F(w) with interpolated values.

        Args:
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz").
                Case-insensitive.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            fontsize: Legend and title fontsize

        Returns: |matplotlib-Figure|
        """
        what_list = ["a2f", "lambda"]
        nrows, ncols = len(what_list), 1
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = np.array(ax_list).ravel()

        styles = dict(
            qcoarse={"linestyle": "--", "color": "b"},
            qintp={"linestyle": "-", "color": "r"},
        )

        for ix, (ax, what) in enumerate(zip(ax_list, what_list)):
            for qsamp in ["qcoarse", "qintp"]:
                a2f = self.get_a2f_qsamp(qsamp)
                a2f.plot(what=what, ax=ax, units=units, ylims=ylims, fontsize=fontsize,
                         label=qsamp if ix == 0 else None,
                         show=False, **styles[qsamp])

        return fig

    @add_fig_kwargs
    def plot_with_a2f(self, what="gamma", units="eV", qsamp="qcoarse", phdos=None, ylims=None, **kwargs) -> Figure:
        """
        Plot phonon bands with lambda(q, nu) + a2F(w) + phonon DOS.

        Args:
            what: ``lambda`` for eph coupling strength, ``gamma`` for phonon linewidths.
            units: Units for phonon plots. Possible values in ("eV", "meV", "Ha", "cm-1", "Thz"). Case-insensitive.
            qsamp:
            phdos: |PhononDos| object. Used to plot the PhononDos on the right.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used

        Returns: |matplotlib-Figure|
        """
        # Max three additional axes with [a2F, a2F_tr, DOS]
        ncols = 2
        width_ratios = [1, 0.2]
        if self.has_a2ftr:
            ncols += 1
            width_ratios.append(0.2)

        if phdos is not None:
            phdos = PhononDos.as_phdos(phdos)
            ncols += 1
            width_ratios.append(0.2)

        # Build grid plot.
        import matplotlib.pyplot as plt
        from matplotlib.gridspec import GridSpec
        fig = plt.figure()
        gspec = GridSpec(1, ncols, width_ratios=width_ratios, wspace=0.05)
        ax_phbands = plt.subplot(gspec[0])

        ax_doses = []
        for i in range(ncols - 1):
            ax = plt.subplot(gspec[i + 1], sharey=ax_phbands)
            ax.grid(True)
            set_axlims(ax, ylims, "y")
            ax_doses.append(ax)

        # Plot phonon bands with markers.
        self.plot(what=what, units=units, ylims=ylims, ax=ax_phbands, show=False)

        # Plot a2F(w)
        a2f = self.get_a2f_qsamp(qsamp)
        ax = ax_doses[0]
        a2f.plot(units=units, exchange_xy=True, ylims=ylims, ax=ax, show=False)
        ax.yaxis.set_ticks_position("right")
        #ax.yaxis.set_label_position("right")
        #ax.tick_params(labelbottom='off')
        ax.set_ylabel("")

        # Plot a2Ftr(w)
        ix = 1
        if self.has_a2ftr:
            ax = ax_doses[ix]
            a2ftr = self.get_a2ftr_qsamp(qsamp)
            self.a2ftr.plot(units=units, exchange_xy=True, ylims=ylims, ax=ax, show=False)
            ax.yaxis.set_ticks_position("right")
            #ax.yaxis.set_label_position("right")
            #ax.tick_params(labelbottom='off')
            ax.set_ylabel("")
            ix += 1

        # Plot DOS g(w)
        if phdos is not None:
            ax = ax_doses[ix]
            phdos.plot_dos_idos(ax=ax, exchange_xy=True, what="d", color="k", linestyle="-")
            ax.yaxis.set_ticks_position("right")
            #ax.yaxis.set_label_position("right")
            #ax.tick_params(labelbottom='off')
            ax.set_xlabel(r"$F(\omega)$")
            #ax.set_ylabel("")

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        Used in abiview.py to get a quick look at the results.
        """
        #yield self.plot(show=False)
        #yield self.plot_eph_strength(show=False)
        #yield self.plot_with_a2f(show=False)

        #for qsamp in ["qcoarse", "qintp"]:
        for qsamp in ["qcoarse",]:
            a2f = self.get_a2f_qsamp(qsamp)
            yield a2f.plot_with_lambda(title="q-sampling: %s (%s)" % (str(a2f.ngqpt), qsamp), show=False)

        #yield self.plot_nuterms(show=False)
        #yield self.plot_a2(show=False)
        #yield self.plot_tc_vs_mustar(show=False)

        #if self.has_a2ftr:
        #    ncfile.a2ftr.plot();

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("ncfile = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(ncfile)"),
            nbv.new_code_cell("ncfile.ebands.plot();"),
            nbv.new_code_cell("ncfile.plot();"),
            #nbv.new_code_cell("ncfile.plot_phlinewidths();"),
            nbv.new_code_cell("ncfile.plot_with_a2f();"),
            nbv.new_code_cell("ncfile.a2f.plot();"),
        ])

        if self.has_a2ftr:
            nb.cells.extend([
                nbv.new_code_cell("ncfile.a2ftr.plot();"),
                #nbv.new_code_cell("ncfile.plot_with_a2ftr();"),
            ])

        return self._write_nb_nbpath(nb, nbpath)


class A2fRobot(Robot, RobotWithEbands, RobotWithPhbands):
    """
    This robot analyzes the results contained in multiple A2F.nc files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: A2fRobot
    """
    #TODO: Method to plot the convergence of DOS(e_F)
    EXT = "A2F"

    linestyle_qsamp = dict(qcoarse="--", qintp="-")
    marker_qsamp = dict(qcoarse="^", qintp="o")

    #all_qsamps = ["qcoarse", "qintp"]
    all_qsamps = ["qcoarse",]

    def get_dataframe(self, abspath=False, with_geo=False, with_params=True, funcs=None) -> pd.DataFrame:
        """
        Build and return a |pandas-DataFrame| with the most important results.

        Args:
            abspath: True if paths in index should be absolute. Default: Relative to getcwd().
            with_geo: True if structure info should be added to the dataframe
            funcs: Function or list of functions to execute to add more data to the DataFrame.
                Each function receives a :class:`A2fFile` object and returns a tuple (key, value)
                where key is a string with the name of column and value is the value to be inserted.
            with_params: False to exclude calculation parameters from the dataframe.

        Return: |pandas-DataFrame|
        """
        rows, row_names = [], []
        for i, (label, ncfile) in enumerate(self.items()):
            row_names.append(label)
            d = OrderedDict()

            for qsamp in self.all_qsamps:
                a2f = ncfile.get_a2f_qsamp(qsamp)
                d["lambda_" + qsamp] = a2f.lambda_iso
                d["omegalog_" + qsamp] = a2f.omega_log

                # Add transport properties.
                if ncfile.has_a2ftr:
                    for qsamp in self.all_qsamps:
                        a2ftr = ncfile.get_a2ftr_qsamp(qsamp)
                        d["lambdatr_avg_" + qsamp] = a2f.lambda_tr

            # Add info on structure.
            if with_geo:
                d.update(ncfile.structure.get_dict4pandas(with_spglib=True))

            # Add convergence parameters
            if with_params:
                d.update(ncfile.params)

            # Execute functions.
            if funcs is not None: d.update(self._exec_funcs(funcs, ncfile))
            rows.append(d)

        row_names = row_names if not abspath else self._to_relpaths(row_names)
        return pd.DataFrame(rows, index=row_names, columns=list(rows[0].keys()))

    @add_fig_kwargs
    def plot_lambda_convergence(self, what="lambda", sortby=None, hue=None, ylims=None, fontsize=8,
                                colormap="jet", **kwargs) -> Figure:
        """
        Plot the convergence of the lambda(q, nu) parameters wrt to the ``sortby`` parameter.

        Args:
            what: "lambda" for eph strength, gamma for phonon linewidths.
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            ylims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used
            fontsize: Legend and title fontsize.
            colormap: matplotlib color map.

        Returns: |matplotlib-Figure|
        """
        # Build (1, ngroups) grid plot.
        if hue is None:
            labels_ncfiles_params = self.sortby(sortby, unpack=False)
            nrows, ncols = 1, 1
        else:
            groups = self.group_and_sortby(hue, sortby)
            nrows, ncols = 1, len(groups)

        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)
        cmap = plt.get_cmap(colormap)

        if hue is None:
            # Plot all results on the same figure with different color.
            for i, (label, ncfile, param) in enumerate(labels_ncfiles_params):
                ncfile.plot_eph_strength(what_list=what,
                        ax_list=[ax_mat[0, 0]],
                        ylims=ylims,
                        label=self.sortby_label(sortby, param),
                        color=cmap(i / len(self)), fontsize=fontsize,
                        show=False,
                        )
        else:
            # ngroup figures
            for ig, g in enumerate(groups):
                ax = ax_mat[0, ig]
                label = "%s: %s" % (self._get_label(hue), g.hvalue)
                for ifile, ncfile in enumerate(g.abifiles):
                    ncfile.plot_eph_strength(what_list=what,
                        ax_list=[ax],
                        ylims=ylims,
                        label=label,
                        color=cmap(ifile / len(g)), fontsize=fontsize,
                        show=False,
                        )
                if ig != 0:
                    set_visible(ax, False, "ylabel")

        return fig

    @add_fig_kwargs
    def plot_a2f_convergence(self, sortby=None, hue=None, qsamps="all", xlims=None,
                            fontsize=8, colormap="jet", **kwargs) -> Figure:
        """
        Plot the convergence of the Eliashberg function wrt to the ``sortby`` parameter.

        Args:
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            qsamps:
            xlims: Set the data limits for the x-axis. Accept tuple e.g. ``(left, right)``
                   or scalar e.g. ``left``. If left (right) is None, default values are used.
            fontsize: Legend and title fontsize.
            colormap: matplotlib color map.

        Returns: |matplotlib-Figure|
        """
        qsamps = self.all_qsamps if qsamps == "all" else list_strings(qsamps)
        #qsamps = ["qcoarse"]

        # Build (2, ngroups) grid plot.
        if hue is None:
            labels_ncfiles_params = self.sortby(sortby, unpack=False)
            nrows, ncols = len(qsamps), 1
        else:
            groups = self.group_and_sortby(hue, sortby)
            nrows, ncols = len(qsamps), len(groups)

        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)
        cmap = plt.get_cmap(colormap)

        for i, qsamp in enumerate(qsamps):
            if hue is None:
                ax = ax_mat[i, 0]
                for j, (label, ncfile, param) in enumerate(labels_ncfiles_params):
                    ncfile.get_a2f_qsamp(qsamp).plot(what="a2f", ax=ax,
                       label=self.sortby_label(sortby, param) + " " + qsamp,
                       color=cmap(j / len(self)), fontsize=fontsize,
                       linestyle=self.linestyle_qsamp[qsamp],
                       show=False,
                    )
                set_axlims(ax, xlims, "x")
            else:
                for ig, g in enumerate(groups):
                    ax = ax_mat[i, ig]
                    label = "%s: %s" % (self._get_label(hue), g.hvalue) + " " + qsamp
                    for ncfile in g.abifiles:
                        ncfile.get_a2f_qsamp(qsamp).plot(what="a2f", ax=ax,
                            label=label,
                            color=cmap(ig / len(g)), fontsize=fontsize,
                            linestyle=self.linestyle_qsamp[qsamp],
                            show=False,
                        )
                    set_axlims(ax, xlims, "x")
                    if ig != 0:
                        set_visible(ax, False, "ylabel")

                if i != len(qsamps) - 1:
                    set_visible(ax, False, "xlabel")

        return fig

    @add_fig_kwargs
    def plot_a2fdata_convergence(self, sortby=None, hue=None, qsamps="all", what_list=("lambda_iso", "omega_log"),
                                 fontsize=8, **kwargs) -> Figure:
        """
        Plot the convergence of the isotropic lambda and omega_log wrt the ``sortby`` parameter.

        Args:
            sortby: Define the convergence parameter, sort files and produce plot labels.
                Can be None, string or function. If None, no sorting is performed.
                If string and not empty it's assumed that the abifile has an attribute
                with the same name and `getattr` is invoked.
                If callable, the output of sortby(abifile) is used.
            hue: Variable that define subsets of the data, which will be drawn on separate lines.
                Accepts callable or string
                If string, it's assumed that the abifile has an attribute with the same name and getattr is invoked.
                If callable, the output of hue(abifile) is used.
            qsamps:
            what_list:
            fontsize: Legend and title fontsize.

        Returns: |matplotlib-Figure|
        """
        what_list = list_strings(what_list)

        # Build grid with (n, 1) plots.
        nrows, ncols = len(what_list), 1
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=False, squeeze=False)
        ax_list = np.array(ax_list).ravel()

        if hue is None:
            labels, ncfiles, params = self.sortby(sortby, unpack=True)
        else:
            groups = self.group_and_sortby(hue, sortby)

        qsamps = self.all_qsamps if qsamps == "all" else list_strings(qsamps)
        marker = kwargs.pop("marker", "o")

        for ix, (ax, what) in enumerate(zip(ax_list, what_list)):
            #ax.set_title(what, fontsize=fontsize)
            if hue is None:
                params_are_string = duck.is_string(params[0])
                xvals = params if not params_are_string else range(len(params))
                l = None
                for iq, qsamp in enumerate(qsamps):
                    a2f_list = [ncfile.get_a2f_qsamp(qsamp) for ncfile in ncfiles]
                    yvals = [getattr(a2f, what) for a2f in a2f_list]
                    l = ax.plot(xvals, yvals,
                                marker=self.marker_qsamp[qsamp],
                                linestyle=self.linestyle_qsamp[qsamp],
                                color=None if iq == 0 else l[0].get_color(),
                                )
                    if params_are_string:
                        ax.set_xticks(xvals)
                        ax.set_xticklabels(params, fontsize=fontsize)
            else:
                for g in groups:
                    for iq, qsamp in enumerate(qsamps):
                        a2f_list = [ncfile.get_a2f_qsamp(qsamp) for ncfile in g.abifiles]
                        yvals = [getattr(a2f, what) for a2f in a2f_list]
                        label = "%s: %s" % (self._get_label(hue), g.hvalue) if iq == 0 else None
                        l = ax.plot(g.xvalues, yvals, label=label,
                                    marker=self.marker_qsamp[qsamp],
                                    linestyle=self.linestyle_qsamp[qsamp],
                                    color=None if iq == 0 else l[0].get_color(),
                                    )

            ax.grid(True)
            ax.set_ylabel(_LATEX_LABELS[what])
            if ix == len(what_list) - 1:
                ax.set_xlabel("%s" % self._get_label(sortby))
                if sortby is None: rotate_ticklabels(ax, 15)
            if hue is not None:
                ax.legend(loc="best", fontsize=fontsize, shadow=True)

        return fig

    @add_fig_kwargs
    def gridplot_a2f(self, xlims=None, fontsize=8, sharex=True, sharey=True, **kwargs) -> Figure:
        """
        Plot grid with a2F(w) and lambda(w) for all files treated by the robot.

        Args:
            xlims: Set the data limits for the y-axis. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If left (right) is None, default values are used
            sharex, sharey: True to share x- and y-axis.
            fontsize: Legend and title fontsize
        """
        return self._gridplot_a2f_what("a2f", xlims=xlims, fontsize=fontsize, sharex=sharex, sharey=sharey, **kwargs)

    #@add_fig_kwargs
    #def gridplot_a2ftr(self, xlims=None, fontsize=8, sharex=True, sharey=True, **kwargs):
    #    return self._gridplot_a2f_what("a2ftr", xlims=xlims, fontsize=fontsize, sharex=sharex, sharey=sharey, **kwargs)

    def _gridplot_a2f_what(self, what, qsamps="all", xlims=None, fontsize=8, sharex=True, sharey=True, **kwargs):
        """Internal method to plot a2F or a2f_tr"""
        nrows, ncols, nplots = 1, 1, len(self)
        if nplots > 1:
            ncols = 2
            nrows = nplots // ncols + nplots % ncols

        # Build grid plot
        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=sharex, sharey=sharey, squeeze=False)
        ax_list = ax_list.ravel()
        # don't show the last ax if nplots is odd.
        if nplots % ncols != 0: ax_list[-1].axis("off")

        qsamps = self.all_qsamps if qsamps == "all" else list_strings(qsamps)
        #qsamps = ["qcoarse"]
        for qsamp in qsamps:
            if what == "a2f":
                a2f_list = [ncfile.get_a2f_qsamp(qsamp) for ncfile in self.abifiles]
            elif what == "a2ftr":
                a2f_list = self.get_a2ftr_qsamp(qsamp)
            else:
                raise ValueError("Invalid value for what: `%s`" % what)

            a2f_list = [ncfile.get_a2f_qsamp(qsamp) for ncfile in self.abifiles]

            for i, (a2f, ax, title) in enumerate(zip(a2f_list, ax_list, self.keys())):
                irow, icol = divmod(i, ncols)
                # FIXME: Twinx is problematic
                a2f.plot_with_lambda(ax=ax, show=False,
                                     linestyle=self.linestyle_qsamp[qsamp],
                                     )

                set_axlims(ax, xlims, "x")
                ax.set_title(title, fontsize=fontsize)
                if (irow, icol) != (0, 0):
                    set_visible(ax, False, "ylabel")
                if irow != nrows - 1:
                    set_visible(ax, False, "xlabel")

        return fig

    #@add_fig_kwargs
    #def plot_a2ftr_convergence(self, sortby=None, qsamps="all", ax=None, xlims=None,
    #                           fontsize=8, colormap="jet", **kwargs):
    #    qsamps = self.all_qsamps if qsamps == "all" else list_strings(qsamps)
    #    ax, fig, plt = get_ax_fig_plt(ax=ax)
    #    cmap = plt.get_cmap(colormap)
    #    for i, (label, ncfile, param) in enumerate(self.sortby(sortby)):
    #        for qsamp in qsamps:
    #            ncfile.get_a2ftr_qsamp(qsamp).plot(
    #                    ax=ax=ax,
    #                    label=self.sortby_label(sortby, param),
    #                    color=cmap(i / len(self)),
    #                    show=False,
    #                    )
    #    set_axlims(ax, xlims, "x")
    #    return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        #yield self.plot_lambda_convergence(show=False)
        yield self.plot_a2f_convergence(show=False)
        yield self.plot_a2fdata_convergence(show=False)
        yield self.gridplot_a2f(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.A2fRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            nbv.new_code_cell("data = robot.get_dataframe()\ndata"),
            nbv.new_code_cell("robot.plot_lambda_convergence();"),
            nbv.new_code_cell("robot.plot_a2f_convergence();"),
        ])

        if all(ncf.has_a2ftr for ncf in self.abifiles):
            nb.cells.extend([
                nbv.new_code_cell("robot.plot_a2ftr_convergence();"),
            ])

        # Mixins.
        nb.cells.extend(self.get_baserobot_code_cells())
        nb.cells.extend(self.get_ebands_code_cells())
        nb.cells.extend(self.get_phbands_code_cells())

        return self._write_nb_nbpath(nb, nbpath)


class A2fReader(BaseEphReader):
    """
    Reads data from the EPH.nc file and constructs objects.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: A2fReader
    """
    def read_edos(self) -> ElectronDos:
        """
        Read the |ElectronDos| used to compute EPH quantities.
        """
        mesh = self.read_value("edos_mesh") * units.Ha_to_eV
        # [nsppol+1, nw] arrays with TOT_DOS, Spin_up, Spin_down in a.u.
        var = self.read_variable("edos_dos")
        if var.shape[0] == 3:
            # Spin polarized. Extract up-down components.
            spin_dos = var[1:, :] / units.Ha_to_eV
        else:
            # Spin unpolarized. Extract Tot DOS
            spin_dos = var[0, :] / units.Ha_to_eV

        #spin_idos = self.read_variable("edos_idos")[1:, :] / units.Ha_to_eV
        nelect = self.read_value("number_of_electrons")
        fermie = self.read_value("fermi_energy") * units.Ha_to_eV

        return ElectronDos(mesh, spin_dos, nelect, fermie=fermie)

    def read_phbands_qpath(self) -> PhononBands:
        """
        Read and return a |PhononBands| object with frequencies computed along the q-path.
        """
        structure = self.read_structure()

        # Build the list of q-points
        qpoints = Kpath(structure.reciprocal_lattice,
                        frac_coords=self.read_value("qpath"),
                        weights=None, names=None, ksampling=None)

        #nctkarr_t('phfreq_qpath', "dp", "natom3, nqpath"),&
        phfreqs = self.read_value("phfreq_qpath") * units.Ha_to_eV
        phdispl_cart = self.read_value("phdispl_cart_qpath", cmode="c") * units.bohr_to_ang

        linewidths = self.read_phgamma_qpath()
        if self.read_nsppol() == 2:
            # We have spin-resolved linewidths, sum over spins here.
            linewidths = linewidths.sum(axis=0)

        amu_list = self.read_value("atomic_mass_units", default=None)
        if amu_list is not None:
            atom_species = self.read_value("atomic_numbers")
            amu = {at: a for at, a in zip(atom_species, amu_list)}
        else:
            raise ValueError("atomic_mass_units is not present!")
            amu = None

        return PhononBands(structure=structure,
                           qpoints=qpoints,
                           phfreqs=phfreqs,
                           phdispl_cart=phdispl_cart,
                           non_anal_ph=None,
                           amu=amu,
                           linewidths=linewidths,
                           )

    def read_phlambda_qpath(self, sum_spin=True):
        """
        Reads the EPH coupling strength *interpolated* along the q-path.

        Return:
            |numpy-array| with shape [nqpath, natom3] if not sum_spin else [nsppol, nqpath, natom3]
        """
        vals = self.read_value("phlambda_qpath")
        return vals if not sum_spin else vals.sum(axis=0)

    def read_phgamma_qpath(self, sum_spin=True):
        """
        Reads the phonon linewidths (eV) *interpolated* along the q-path.

        Return:
            |numpy-array| with shape [nqpath, natom3] if not sum_spin else [nsppol, nqpath, natom3]
        """
        vals = self.read_value("phgamma_qpath") * units.Ha_to_eV
        return vals if not sum_spin else vals.sum(axis=0)

    def read_a2f(self, qsamp):
        """
        Read and return the Eliashberg function :class:`A2F`.
        """
        assert qsamp in ("qcoarse", "qintp")
        mesh = self.read_value("a2f_mesh_" + qsamp) * units.Ha_to_eV
        # C shape [nsppol, natom + 1, nomega]
        data = self.read_value("a2f_values_" + qsamp) # * 0.25
        values_spin = data[:, 0, :].copy()
        values_spin_nu = data[:, 1:, :].copy()

        # Extract q-mesh and meta variables.
        ngqpt = self.ngqpt if qsamp == "qcoarse" else self.ph_ngqpt
        meta = {k: self.common_eph_params[k] for k in
                ["eph_intmeth", "eph_fsewin", "eph_fsmear", "eph_extrael", "eph_fermie"]}

        return A2f(mesh, values_spin, values_spin_nu, ngqpt, meta)

    #def read_a2ftr(self, qsamp):
    #    """Read and return the Eliashberg transport spectral function a2F_tr(w, x, x')."""
    #    assert qsamp in ("qcoarse", "qintp")
    #    mesh = self.read_value("a2ftr_mesh_" + qsamp) * units.Ha_to_eV
    #    # Transpose tensor components F --> C
    #    vals_in = self.read_value("a2ftr_in_" + qsamp)
    #    vals_out = self.read_value("a2ftr_out_" + qsamp)
    #    return A2ftr(mesh=mesh, vals_in, vals_out)

    #def read_phgamma_ibz_data(self):
    #     ! linewidths in IBZ
    #     nctkarr_t('qibz', "dp", "number_of_reduced_dimensions, nqibz"), &
    #     nctkarr_t('wtq', "dp", "nqibz"), &
    #     nctkarr_t('phfreq_qibz', "dp", "natom3, nqibz"), &
    #     nctkarr_t('phdispl_cart_qibz', "dp", "two, natom3, natom3, nqibz"), &
    #     nctkarr_t('phgamma_qibz', "dp", "natom3, nqibz, number_of_spins"), &
    #     nctkarr_t('phlambda_qibz', "dp", "natom3, nqibz, number_of_spins") &



# flake8: noqa
from pymatgen.core import __version__ as pmg_version
from pymatgen.io.abinit.abitimer import AbinitTimerParser, AbinitTimerSection


if pmg_version < '2023.7.10':
    from pymatgen.io.abinit.abitimer import AbinitTimerParserError as AbinitTimerParseError
else:
    from pymatgen.io.abinit.abitimer import AbinitTimerParseError



#!/usr/bin/env python

import sys

from abipy.core.structure import Structure
from abipy.htc.base_models import MongoConnector
from abipy.htc.protocol import Protocol
from abipy.htc.structure_models import StructureData
from abipy.htc.pseudos_models import PseudoSpecs
from abipy.htc.gs_flow_models import EbandsFlowModelWithParams, EbandsFlowModelWithProtocol
from abipy.htc.dfpt_flow_models import PhononFlowModelWithParams, PhononFlowModelWithInput

# Explicit options.
#mng_connector = MongoConnector(host="mongo.pcpm.ucl.ac.be",
#                               db_name="abipy_GM",
#                               collection_name="ebands",
#                               username="gmatteo",
#                               password="Ts5qb/UtB.22",
#                               )

collection_name = "ebands"

# Take connection options from ~/.abinit/abipy/config.yml configuration file.
#mng_connector = MongoConnector.from_abipy_config(collection_name=collection_name)

# Connect to localhost for testing/debugging purposes
mng_connector = MongoConnector.for_localhost(collection_name=collection_name)
print(mng_connector)

collection = mng_connector.get_collection()

# Pseudopotential specifications.
# Note that we need to specify the repo name plus the table_name
#
pseudos_specs = PseudoSpecs.from_repo_table_name("ONCVPSP-PBE-SR-PDv0.4", "standard")
#pseudos_specs = PseudoSpecs.from_repo_table_name("ONCVPSP-PBE-FR-PDv0.4", "standard")

protocol = Protocol.from_name("NC-PBE-SR-PDv0.4.yml")
print(protocol)


def build_collection(drop=False):
    """
    Fill a MongoDB collection with FlowModel instances defining the kind of
    calculation we want to perform and the schema used to store the output results.
    This step is mandatory before starting the AbipyWorker.
    """
    if drop:
        mng_connector.drop_collection(ask_for_confirmation=False)
    else:
        print(f"Adding documents to collection: {collection_name}")

    # Generate list of structures.
    quick = True
    #quick = False
    if quick:
        from abipy.data.ucells import structure_from_ucell
        structures = [structure_from_ucell(name) for name in ("Si", "Si-shifted")]

    else:
        structures = [Structure.from_mpid(mpid) for mpid in (
            #"mp-149",   # Si2
            #"mp-2172",  # AlAs F-43m
            #"mp-2534",  # GaAs F-43m
            "mp-32",     # Ge-dia
            "mp-149",    # Si-dia
            "mp-2534",   # GaAs-zb
            "mp-406",    # CdTe-zb
            "mp-2691",   # CdSe-zb
            "mp-2172",   # AlAs-zb
            "mp-2176",   # ZnTe-zb
            "mp-8062",   # SiC-zb
            "mp-2469",   # CdS-zb
            "mp-1550",   # AlP-zb
            "mp-1190",   # ZnSe-zb
            "mp-2657",   # TiO$_2$-t
            "mp-5229",   # SrTiO$_3$-sc
            "mp-804",    # GaN-w
            "mp-830",    # GaN-zb
            "mp-2133",   # ZnO-w
            "mp-856",    # SnO$_2$-t
            "mp-10695",  # ZnS-zb
            "mp-1342",   # BaO-rs
            "mp-2472",   # SrO-rs
            "mp-66",     # C-dia
            "mp-661",    # AlN-w
            "mp-1639",   # BN-zb
            "mp-2605",   # CaO-rs
            "mp-1960",   # Li$_2$O
            "mp-1265",   # MgO-rs
            "mp-6947",   # SiO$_2$-t
            "mp-2542",   # BeO-w
            "mp-1138",   # LiF-rs
        )]

    # Get pseudopotential tables with hints.
    pseudos = pseudos_specs.get_pseudos()

    # Build list of FlowModels for each structure.

    #flow_model_cls = FlowModelWithProto.get_class_from_name()

    flow_models = []
    for i, structure in enumerate(structures):
        m = EbandsFlowModelWithProtocol.from_structure_and_protocol(structure, protocol)

        #in_structure_data = StructureData.from_structure(structure)

        #m = EbandsFlowModelWithParams(in_structure_data=in_structure_data,
        #                              pseudos_specs=pseudos_specs,
        #                              spin_mode="unpolarized",
        #                              #kppa=300,
        #                              #kppa=1000,
        #                              ndivsm=2,
        #                              dos_kppa=None,
        #                              with_gsr=True,  # Save GSR.nc in DB
        #                              with_out=True,  # Save run.abo in DB
        #                              with_log=True,  # Save run.log in DB
        #                              )

        #m = EbandsFlowModelWithInputs(in_structure_data=in_structure_data,
        #                              pseudos_specs=pseudos_specs,
        #                              scf_input=...,
        #                              nscf_input=..,
        #                              with_gsr=True,
        #                              with_out=True,
        #                              with_log=True,
        #                              )


        #m = RelaxFlowModel(in_structure_data=in_structure_data,
        #                   spin_mode="unpolarized",
        #                   pseudos_specs=pseudos_specs,
        #                   kppa=300)

        #accuracy = "normal"
        #scf_input = make_scf_input(structure, pseudos, accuracy)
        #m = PhononFlowModelWithInput(in_structure_data=in_structure_data,
        #                             pseudos_specs=pseudos_specs,
        #                             scf_input=scf_input,
        #                             with_becs=False,
        #                             with_quad=False,
        #                             with_gsden=True,
        #                             wtih_gspot=True,
        #                             with_dvdb=True,
        #                             )

        flow_models.append(m)

    # Inser FlowModels in collection
    mng_connector.insert_flow_models(flow_models, verbose=1)


def make_scf_input(structure, pseudos, accuracy, paral_kgb=0):
    """
    This function constructs the input file for the GS calculation:
    """
    # Crystalline AlAs: computation of the second derivative of the total energy
    from abipy import abilab
    #structure = abidata.structure_from_ucell("AlAs")
    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    #num_valence_electrons = gs_inp.structure.num_valence_electrons(pseudos)
    num_valence_electrons = gs_inp.num_valence_electrons
    nsppol = 2
    nband = num_valence_electrons // nsppol + 4
    nband += nband % 2

    gs_inp.set_vars(
        nband=nband,
        #nband=4,
        #ecut=2.0,
        ngkpt=[4, 4, 4],
        #shiftk=[0.0, 0.0, 0.5,   # This gives the usual fcc Monkhorst-Pack grid
        #        0.0, 0.5, 0.0,
        #        0.5, 0.0, 0.0,
        #        0.5, 0.5, 0.5],
        shiftk=[0, 0, 0],
        paral_kgb=paral_kgb,
        tolvrs=1.0e-8,
        diemac=9.0,
    )

    gs_inp.set_cutoffs_for_accuracy(accuracy)
    #gs_inp.set_auto_scf_nband(nsppol=1, nspinor=1, nspden=1, occopt, tsmear)
    return gs_inp


#def aggregate():
#    collection = mng_connector.get_collection()
#    df = EbandsFlowModelWithParams.mng_aggregate_in_structures(collection)
#    from abipy.abilab import print_dataframe
#    print_dataframe(df)



if __name__ == "__main__":
    try:
        command = sys.argv[1]
    except IndexError:
        print("Available commands: `build`, `add`")
        sys.exit(1)

    if command == "build":
        build_collection(drop=True)
    elif command == "add":
        build_collection(drop=False)
    elif command == "aggregate":
        aggregate()
    else:
        raise ValueError(f"Invalid command: {command}")



# Copyright (c) Pymatgen Development Team.
# Distributed under the terms of the MIT License.
"""
Most features of this module has been moved to monty. Please refer to
monty.json and monty.serialization documentation.
"""
from __future__ import annotations

import functools
import json
import pickle
import json

from typing import Any
from monty.json import MontyDecoder, MontyEncoder
from pymatgen.core.periodic_table import Element


def pmg_serialize(method):
    """
    Decorator for methods that add MSON serializations keys
    to the dictionary. See documentation of MSON for more details
    """

    @functools.wraps(method)
    def wrapper(*args, **kwargs):
        self = args[0]
        d = method(*args, **kwargs)
        # Add @module and @class
        d["@module"] = type(self).__module__
        d["@class"] = type(self).__name__
        return d

    return wrapper


def json_pretty_dump(obj: Any, filename: str) -> None:
    """
    Serialize obj as a JSON formatted stream to the given filename (
    pretty printing version)
    """
    with open(filename, "w") as fh:
        json.dump(obj, fh, indent=4, sort_keys=4)


class PmgPickler(pickle.Pickler):
    """
    Persistence of External Objects as described in section 12.1.5.1 of
    https://docs.python.org/3/library/pickle.html
    """

    def persistent_id(self, obj: Any):
        """Instead of pickling as a regular class instance, we emit a
        persistent ID."""
        if isinstance(obj, Element):
            # Here, our persistent ID is simply a tuple, containing a tag and
            # a key
            return type(obj).__name__, obj.symbol
        # If obj does not have a persistent ID, return None. This means obj
        # needs to be pickled as usual.
        return None


class PmgUnpickler(pickle.Unpickler):
    """
    Persistence of External Objects as described in section 12.1.5.1 of
    https://docs.python.org/3/library/pickle.html
    """

    def persistent_load(self, pid):
        """
        This method is invoked whenever a persistent ID is encountered.
        Here, pid is the tuple returned by PmgPickler.
        """
        try:
            type_tag, key_id = pid
        except Exception:
            # Sometimes we get a string such as ('Element', u'C') instead
            # of a real tuple. Use ast to evaluate the expression (much safer
            # than eval).
            import ast
            type_tag, key_id = ast.literal_eval(pid)

        if type_tag == "Element":
            return Element(key_id)

        # Always raises an error if you cannot return the correct object.
        # Otherwise, the unpickler will think None is the object referenced
        # by the persistent ID.
        raise pickle.UnpicklingError(f"unsupported persistent object with pid {pid}")


def pmg_pickle_load(filobj, **kwargs) -> Any:
    """
    Loads a pickle file and deserialize it with PmgUnpickler.

    Args:
        filobj: File-like object
        **kwargs: Any of the keyword arguments supported by PmgUnpickler
    Returns:
        Deserialized object.
    """
    return PmgUnpickler(filobj, **kwargs).load()


def pmg_pickle_dump(obj: Any, filobj, **kwargs):
    """
    Dump an object to a pickle file using PmgPickler.

    Args:
        obj: Object to dump.
        fileobj: File-like object
        **kwargs: Any of the keyword arguments supported by PmgPickler
    """
    return PmgPickler(filobj, **kwargs).dump(obj)


def mjson_load(filepath: str, **kwargs) -> Any:
    """
    Read JSON file in MSONable format with MontyDecoder.
    Return dict with python object.
    """
    with open(filepath, "rt") as fh:
        return json.load(fh, cls=MontyDecoder, **kwargs)


def mjson_loads(string: str, **kwargs) -> Any:
    """
    Read JSON string in MSONable format with MontyDecoder.
    Return python object
    """
    return json.loads(string, cls=MontyDecoder, **kwargs)


def mjson_write(d: dict, filepath: str, **kwargs) -> None:
    """
    Write dictionary d to filepath in JSON format using MontyDecoder
    """
    with open(filepath, "wt") as fh:
        json.dump(d, fh, cls=MontyEncoder, **kwargs)



#!/usr/bin/env python
r"""
Speed of Sound
==============

This example shows how to compute the speed of sound by fitting phonon frequencies
along selected directions by linear least-squares fit.
For a command line interface use:

.. code-block:: bash

    abiview.py ddb_vs DDB_FILE
"""

#%%
# Initialize object from DDB file.

import os
import abipy.data as abidata

from abipy import abilab
from abipy.dfpt.vsound import SoundVelocity

ddb_path = os.path.join(abidata.dirpath, "refs", "si_sound_vel", "Si_DDB")
sv = SoundVelocity.from_ddb(ddb_path)

#%%
# Get pandas dataframe with results.

df = sv.get_dataframe()
abilab.print_dataframe(df)

#%%
# Plot fit with matplotlib

sv.plot()

#%%
# Plot fit with plotly

sv.plotly(template="plotly_dark")



"""
Objects used to extract and plot results from output files in text format.
"""
from __future__ import annotations

import os
import numpy as np
import pandas as pd

from collections import OrderedDict
from io import StringIO
from typing import Union
from monty.string import is_string, marquee
from monty.functools import lazy_property
from monty.termcolor import cprint
from pymatgen.core.units import bohr_to_ang
from abipy.core.symmetries import AbinitSpaceGroup
from abipy.core.structure import Structure, dataframes_from_structures
from abipy.core.kpoints import has_timrev_from_kptopt
from abipy.core.mixins import TextFile, AbinitNcFile, NotebookWriter
from abipy.tools.typing import Figure
from abipy.abio.inputs import GEOVARS
from abipy.abio.timer import AbinitTimerParser
from abipy.abio.robots import Robot
#from abipy.tools.plotting import (set_axlims, add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt,
#    get_ax3d_fig_plt, rotate_ticklabels, set_visible, plot_unit_cell, set_ax_xylabels, get_figs_plotly,
#    get_fig_plotly, add_plotly_fig_kwargs, PlotlyRowColDesc, plotly_klabels, plotly_set_lims)
from abipy.flowtk import EventsParser, NetcdfReader, GroundStateScfCycle, D2DEScfCycle


class AbinitTextFile(TextFile):
    """
    Base class for the ABINIT main output files and log files.
    """
    @property
    def events(self) -> list:
        """
        List of ABINIT events reported in the file.
        """
        # Parse the file the first time the property is accessed or when mtime is changed.
        stat = os.stat(self.filepath)
        if stat.st_mtime != self._last_mtime or not hasattr(self, "_events"):
            self._events = EventsParser().parse(self.filepath)
        return self._events

    def get_timer(self) -> AbinitTimerParser:
        """
        Timer data.
        """
        timer = AbinitTimerParser()
        timer.parse(self.filepath)
        return timer


class AbinitLogFile(AbinitTextFile, NotebookWriter):
    """
    Class representing the ABINIT log file.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: AbinitLogFile
    """

    def to_string(self, verbose=0) -> str:
        return str(self.events)

    def plot(self, **kwargs):
        """Empty placeholder."""
        return None

    #@add_fig_kwargs
    #def plot_mem(self, **kwargs) -> Figure:
    #   return fig

    #@add_fig_kwargs
    #def plot_time(self, **kwargs) -> Figure:
    #   return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield None

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("abilog = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(abilog.events)"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class AbinitOutputFile(AbinitTextFile, NotebookWriter):
    """
    Class representing the main Abinit output file.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: AbinitOutputFile
    """
    # TODO: Extract number of errors and warnings.

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.debug_level = 0
        self._parse()

    def _parse(self) -> None:
        """
        header: String with the input variables
        footer: String with the output variables
        datasets: Dictionary mapping dataset index to list of strings.
        """
        # Get code version and find magic line signaling that the output file is completed.
        self.version, self.run_completed = None, False
        self.overall_cputime, self.overall_walltime = 0.0, 0.0
        self.proc0_cputime, self.proc0_walltime = 0.0, 0.0

        with open(self.filepath, "rt") as fh:
            for line in fh:
                if self.version is None and line.startswith(".Version"):
                    self.version = line.split()[1]

                if line.startswith("- Proc."):
                    #- Proc.   0 individual time (sec): cpu=         25.5  wall=         26.1
                    tokens = line.split()
                    self.proc0_walltime = float(tokens[-1])
                    self.proc0_cputime = float(tokens[-3])

                if line.startswith("+Overall time"):
                    #+Overall time at end (sec) : cpu=         25.5  wall=         26.1
                    tokens = line.split()
                    self.overall_cputime = float(tokens[-3])
                    self.overall_walltime = float(tokens[-1])

                if " Calculation completed." in line:
                    self.run_completed = True

        # Parse header to get important dimensions and variables
        self.header, self.footer, self.datasets = [], [], OrderedDict()
        where = "in_header"

        with open(self.filepath, "rt") as fh:
            for line in fh:
                if "== DATASET" in line:
                    # Save dataset number
                    # == DATASET  1 ==================================================================
                    where = int(line.replace("=", "").split()[-1])
                    assert where not in self.datasets
                    self.datasets[where] = []
                elif "== END DATASET(S) " in line:
                    where = "in_footer"

                if where == "in_header":
                    self.header.append(line)
                elif where == "in_footer":
                    self.footer.append(line)
                else:
                    # dataset number --> lines
                    self.datasets[where].append(line)

        self.header = "".join(self.header)
        if self.debug_level: print("header:\n", self.header)
        # Output files produced in dryrun_mode contain the following line:
        # abinit : before driver, prtvol=0, debugging mode => will skip driver
        self.dryrun_mode = "debugging mode => will skip driver" in self.header
        #print("dryrun_mode:", self.dryrun_mode)

        #if " jdtset " in self.header: raise NotImplementedError("jdtset is not supported")
        #if " udtset " in self.header: raise NotImplementedError("udtset is not supported")

        self.ndtset = len(self.datasets)
        if not self.datasets:
            #raise NotImplementedError("Empty dataset sections.")
            self.ndtset = 1
            self.datasets[1] = "Empty dataset"

        for key, data in self.datasets.items():
            if self.debug_level: print("data")
            self.datasets[key] = "".join(data)
            if self.debug_level: print(self.datasets[key])

        self.footer = "".join(self.footer)
        if self.debug_level: print("footer:\n", self.footer)

        self.initial_vars_global, self.initial_vars_dataset = self._parse_variables("header")
        self.final_vars_global, self.final_vars_dataset = None, None
        if self.run_completed:
            if self.dryrun_mode:
                # footer is not present. Copy values from header.
                self.final_vars_global, self.final_vars_dataset = self.initial_vars_global, self.initial_vars_dataset
            else:
                self.final_vars_global, self.final_vars_dataset = self._parse_variables("footer")

    def _parse_variables(self, what: str):
        vars_global = OrderedDict()
        vars_dataset = OrderedDict([(k, OrderedDict()) for k in self.datasets.keys()])
        #print("keys", vars_dataset.keys())

        lines = getattr(self, what).splitlines()
        if what == "header":
            magic_start = " -outvars: echo values of preprocessed input variables --------"
        elif what == "footer":
            magic_start = " -outvars: echo values of variables after computation  --------"
        else:
            raise ValueError("Invalid value for what: `%s`" % str(what))

        magic_stop = "================================================================================"

        # Select relevant portion with variables.
        for i, line in enumerate(lines):
            if magic_start in line:
                break
        else:
            raise ValueError("Cannot find magic_start line: `%s`\nPerhaps this is not an Abinit output file!" % magic_start)
        lines = lines[i+1:]

        for i, line in enumerate(lines):
            if magic_stop in line:
                break
        else:
            raise ValueError("Cannot find magic_stop line: `%s`\nPerhaps this is not an Abinit output file!" % magic_stop)
        lines = lines[:i]

        # Parse data. Assume format:
        #   timopt          -1
        #    tnons      0.0000000  0.0000000  0.0000000     0.2500000  0.2500000  0.2500000
        #               0.0000000  0.0000000  0.0000000     0.2500000  0.2500000  0.2500000
        def get_dtindex_key_value(line):
            tokens = line.split()
            s, value = tokens[0], " ".join(tokens[1:])
            l = []
            for i, c in enumerate(s[::-1]):
                if c.isalpha():
                    key = s[:len(s)-i]
                    break
                l.append(c)
            else:
                raise ValueError("Cannot find dataset index in token: %s\n" % s)

            #print(line, "\n", l)
            dtindex = None
            if l:
                l.reverse()
                dtindex = int("".join(l))
            return dtindex, key, value

        # (varname, dtindex), [line1, line2 ...]
        stack_var, stack_lines = None, []

        def pop_stack():
            if stack_lines:
                key, dtidx = stack_var
                value = " ".join(stack_lines)
                if dtidx is None:
                    vars_global[key] = value
                else:
                    vars_dataset[dtidx][key] = value

        for line in lines:
            if not line: continue
            # Ignore first char
            line = line[1:].lstrip().rstrip()
            if not line: continue
            #print("line", line)
            if line[0].isalpha():
                pop_stack()
                stack_lines = []
                dtidx, key, value = get_dtindex_key_value(line)
                stack_var = (key, dtidx)
                stack_lines.append(value)
            else:
                stack_lines.append(line)

        pop_stack()

        return vars_global, vars_dataset

    def _get_structures(self, what: str) -> list[Structure]:
        if what == "header":
            vars_global, vars_dataset = self.initial_vars_global, self.initial_vars_dataset
        elif what == "footer":
            vars_global, vars_dataset = self.final_vars_global, self.final_vars_dataset
        else:
            raise ValueError("Invalid value for what: `%s`" % str(what))

        #print("global", vars_global["acell"])
        from abipy.abio.abivars import is_abiunit
        inigeo = {k: vars_global[k] for k in GEOVARS if k in vars_global}

        spgvars = ("spgroup", "symrel", "tnons", "symafm")
        spgd_global = {k: vars_global[k] for k in spgvars if k in vars_global}
        global_kptopt = vars_global.get("kptopt", 1)

        structures = []
        for i in self.datasets:
            # This code breaks down if there are conflicting GEOVARS in globals and dataset.
            d = inigeo.copy()
            d.update({k: vars_dataset[i][k] for k in GEOVARS if k in vars_dataset[i]})

            for key, value in d.items():
                # Must handle possible unit.
                fact = 1.0
                tokens = [t.lower() for t in value.split()]
                if is_abiunit(tokens[-1]):
                    tokens, unit = tokens[:-1], tokens[-1]
                    if unit in ("angstr", "angstrom", "angstroms"):
                        fact = 1.0 / bohr_to_ang
                    elif unit in ("bohr", "bohrs", "au"):
                        fact = 1.0
                    else:
                        raise ValueError("Don't know how to handle unit: %s" % unit)

                s = " ".join(tokens)
                dtype = float if key not in ("ntypat", "typat", "natom") else int
                try:
                    #print(key, s)
                    value = np.fromstring(s, sep=" ", dtype=dtype)
                    #print(key, value)
                    if fact != 1.0: value *= fact # Do not change integer arrays e.g typat!
                    d[key] = value
                except ValueError as exc:
                    print(key, s)
                    raise exc

            if "rprim" not in d and "angdeg" not in d: d["rprim"] = np.eye(3)
            if "natom" in d and d["natom"] == 1 and all(k not in d for k in ("xred", "xcart", "xangst")):
                d["xred"] = np.zeros(3)
            #print(d)
            abistr = Structure.from_abivars(d)

            # Extract Abinit spacegroup.
            spgd = spgd_global.copy()
            spgd.update({k: vars_dataset[i][k] for k in spgvars if k in vars_dataset[i]})

            spgid = int(spgd.get("spgroup", 0))
            if "symrel" not in spgd:
                symrel = np.reshape(np.eye(3, 3, dtype=int), (1, 3, 3))
                spgd["symrel"] = " ".join((str(i) for i in symrel.flatten()))
            else:
                symrel = np.reshape(np.array([int(n) for n in spgd["symrel"].split()], dtype=int), (-1, 3, 3))
            nsym = len(symrel)
            assert nsym == spgd.get("nsym", nsym) #; print(symrel.shape)

            if "tnons" in spgd:
                tnons = np.reshape(np.array([float(t) for t in spgd["tnons"].split()], dtype=float), (nsym, 3))
            else:
                tnons = np.zeros((nsym, 3))

            if "symafm" in spgd:
                symafm = np.array([int(n) for n in spgd["symafm"].split()], dtype=int)
                symafm.shape = (nsym,)
            else:
                symafm = np.ones(nsym, dtype=int)

            try:
                has_timerev = has_timrev_from_kptopt(vars_dataset[i].get("kptopt", global_kptopt))
                abi_spacegroup = AbinitSpaceGroup(spgid, symrel, tnons, symafm, has_timerev, inord="C")
                abistr.set_abi_spacegroup(abi_spacegroup)
            except Exception as exc:
                print("Cannot build AbinitSpaceGroup from the variables reported in file!\n", str(exc))

            structures.append(abistr)

        return structures

    @lazy_property
    def initial_structures(self) -> list[Structure]:
        """List of initial |Structure|."""
        return self._get_structures("header")

    @property
    def has_same_initial_structures(self) -> bool:
        """True if all initial structures are equal."""
        return all(self.initial_structures[0] == s for s in self.initial_structures)

    @lazy_property
    def final_structures(self) -> list[Structure]:
        """List of final |Structure|."""
        if self.run_completed:
            return self._get_structures("footer")
        else:
            cprint("Cannot extract final structures from file.\n %s" % self.filepath, "red")
            return []

    @lazy_property
    def initial_structure(self) -> Structure:
        """
        The |Structure| defined in the output file.

        If the input file contains multiple datasets **AND** the datasets
        have different structures, this property returns None.
        In this case, one has to access the structure of the individual datasets.
        For example:

            self.initial_structures[0]

        gives the structure of the first dataset.
        """
        if not self.has_same_initial_structures:
            print("Datasets have different structures. Returning None. Use initial_structures[0]")
            return None
        return self.initial_structures[0]

    @property
    def has_same_final_structures(self) -> bool:
        """True if all initial structures are equal."""
        return all(self.final_structures[0] == s for s in self.final_structures)

    @lazy_property
    def final_structure(self) -> Union[Structure, None]:
        """
        The |Structure| defined in the output file.

        If the input file contains multiple datasets **AND** the datasets
        have different structures, this property returns None.
        In this case, one has to access the structure of the individual datasets.
        For example:

            self.final_structures[0]

        gives the structure of the first dataset.
        """
        if not self.has_same_final_structures:
            print("Datasets have different structures. Returning None. Use final_structures[0]")
            return None
        return self.final_structures[0]

    def diff_datasets(self, dt_list1, dt_list2, with_params=True, differ="html", dryrun=False):
        """
        Compare datasets
        """
        if not isinstance(dt_list1, (list, tuple)): dt_list1 = [dt_list1]
        if not isinstance(dt_list2, (list, tuple)): dt_list2 = [dt_list2]

        dt_lists = [dt_list1, dt_list2]
        import tempfile
        tmp_names = []
        for i in range(2):
            _, tmpname = tempfile.mkstemp(text=True)
            tmp_names.append(tmpname)
            with open(tmpname, "wt") as fh:
                if with_params: fh.write(self.header)
                for idt in dt_lists[i]:
                    fh.write(self.datasets[idt])
                if with_params: fh.write(self.footer)

        if differ == "html":
            from abipy.tools.devtools import HtmlDiff
            diff = HtmlDiff(tmp_names)
            if dryrun:
                return diff
            else:
                return diff.open_browser()
        else:
            cmd = "%s %s %s" % (differ, tmp_names[0], tmp_names[1])
            if dryrun:
                return cmd
            else:
                return os.system(cmd)

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = ["ndtset: %d, completed: %s" % (self.ndtset, self.run_completed)]
        app = lines.append

        # Different cases depending whether final structures are available
        # and whether structures are equivalent.
        if self.run_completed:
            if self.has_same_final_structures:
                if self.initial_structure != self.final_structure:
                    # Structural relaxation.
                    df = dataframes_from_structures([self.initial_structure, self.final_structure],
                                                    index=["initial", "final"])
                    app("Lattice parameters:")
                    app(str(df.lattice))
                    app("Atomic coordinates:")
                    app(str(df.coords))
                else:
                    # initial == final. Print final structure.
                    app(self.final_structure.to_string(verbose=verbose))
        else:
            # Final structures are not available.
            if self.has_same_initial_structures:
                app(self.initial_structure.to_string(verbose=verbose))
            else:
                df = dataframes_from_structures(self.initial_structures,
                                                index=[i+1 for i in range(self.ndtset)])
                app("Lattice parameters:")
                app(str(df.lattice))
                app("Atomic coordinates:")
                app(str(df.coords))

        # Print dataframe with dimensions.
        df = self.get_dims_spginfo_dataframe(verbose=verbose)
        from abipy.tools.printing import print_dataframe
        strio = StringIO()
        print_dataframe(df, file=strio)
        strio.seek(0)
        app("")
        app(marquee("Dimensions of calculation", mark="="))
        app("".join(strio))

        return "\n".join(lines)

    def get_dims_spginfo_dataframe(self, verbose: int = 0) -> pd.DataFrame:
        """
        Parse the section with the dimensions of the calculation. Return Dataframe.
        """
        dims_dataset, spginfo_dataset = self.get_dims_spginfo_dataset(verbose=verbose)
        rows = []
        for dtind, dims in dims_dataset.items():
            d = OrderedDict()
            d["dataset"] = dtind
            d.update(dims)
            d.update(spginfo_dataset[dtind])
            rows.append(d)

        df = pd.DataFrame(rows, columns=list(rows[0].keys()) if rows else None)
        df = df.set_index('dataset')
        return df

    def get_dims_spginfo_dataset(self, verbose=0) -> tuple[dict, dict]:
        """
        Parse the section with the dimensions of the calculation. Return dictionaries

        Args:
            verbose: Verbosity level.

        Return: (dims_dataset, spginfo_dataset)

            where dims_dataset[i] is an OrderedDict with the dimensions of dataset `i`
            spginfo_dataset[i] is a dictionary with space group information.
        """
        # If single dataset, we have to parse
        #
        #  Symmetries : space group Fd -3 m (#227); Bravais cF (face-center cubic)
        # ================================================================================
        #  Values of the parameters that define the memory need of the present run
        #      intxc =       0    ionmov =       0      iscf =       7    lmnmax =       6
        #      lnmax =       6     mgfft =      18  mpssoang =       3    mqgrid =    3001
        #      natom =       2  nloc_mem =       1    nspden =       1   nspinor =       1
        #     nsppol =       1      nsym =      48    n1xccc =    2501    ntypat =       1
        #     occopt =       1   xclevel =       2
        # -    mband =           8        mffmem =           1         mkmem =          29
        #        mpw =         202          nfft =        5832          nkpt =          29
        # ================================================================================
        # P This job should need less than                       3.389 Mbytes of memory.
        #   Rough estimation (10% accuracy) of disk space for files :
        # _ WF disk file :      0.717 Mbytes ; DEN or POT disk file :      0.046 Mbytes.
        # ================================================================================

        # If multi datasets we have to parse:

        #  DATASET    2 : space group F-4 3 m (#216); Bravais cF (face-center cubic)
        # ================================================================================
        #  Values of the parameters that define the memory need for DATASET  2.
        #      intxc =       0    ionmov =       0      iscf =       7    lmnmax =       2
        #      lnmax =       2     mgfft =      12  mpssoang =       3    mqgrid =    3001
        #      natom =       2  nloc_mem =       1    nspden =       1   nspinor =       1
        #     nsppol =       1      nsym =      24    n1xccc =    2501    ntypat =       2
        #     occopt =       1   xclevel =       1
        # -    mband =          10        mffmem =           1         mkmem =           2
        #        mpw =          69          nfft =        1728          nkpt =           2
        # ================================================================================
        # P This job should need less than                       1.331 Mbytes of memory.
        #   Rough estimation (10% accuracy) of disk space for files :
        # _ WF disk file :      0.023 Mbytes ; DEN or POT disk file :      0.015 Mbytes.
        # ================================================================================

        magic = "Values of the parameters that define the memory need"
        memory_pre = "P This job should need less than"
        magic_exit = "------------- Echo of variables that govern the present computation"
        filesizes_pre = "_ WF disk file :"
        #verbose = 1

        def parse_spgline(line):
            """Parse the line with space group info, return dict."""
            # Could use regular expressions ...
            i = line.find("space group")

            if i == -1:
                # the unit cell is not primitive
                return {}

            spg_str, brav_str = line[i:].replace("space group", "").split(";")
            toks = spg_str.split()
            return {
                "spg_symbol": "".join(toks[:-1]),
                "spg_number": int(toks[-1].replace("(", "").replace(")", "").replace("#", "")),
                "bravais": brav_str.strip(),
            }

        from abipy.tools.numtools import grouper
        dims_dataset, spginfo_dataset = OrderedDict(), OrderedDict()
        inblock = 0
        with open(self.filepath, "rt") as fh:
            for line in fh:
                line = line.strip()
                if verbose > 1: print("inblock:", inblock, " at line:", line)

                if line.startswith(magic_exit): break

                if (not line or line.startswith("===") or line.startswith("---")
                    #or line.startswith("P")
                    or line.startswith("Rough estimation") or line.startswith("PAW method is used")):
                    continue

                if line.startswith("DATASET") or line.startswith("Symmetries :"):
                    # Get dataset index, parse space group and lattice info, init new dims dict.
                    inblock = 1
                    if line.startswith("Symmetries :"):
                        # No multidataset
                        dtindex = 1
                    else:
                        tokens = line.split()
                        dtindex = int(tokens[1])

                    dims_dataset[dtindex] = dims = OrderedDict()
                    spginfo_dataset[dtindex] = parse_spgline(line)
                    continue

                if inblock == 1 and line.startswith(magic):
                    inblock = 2
                    continue

                if inblock == 2:
                    # Lines with data.
                    if line.startswith("For the susceptibility"): continue

                    if line.startswith(memory_pre):
                        dims["mem_per_proc_mb"] = float(line.replace(memory_pre, "").split()[0])
                    elif line.startswith(filesizes_pre):
                        tokens = line.split()
                        mbpos = [i - 1 for i, t in enumerate(tokens) if t.startswith("Mbytes")]
                        assert len(mbpos) == 2
                        dims["wfk_size_mb"] = float(tokens[mbpos[0]])
                        dims["denpot_size_mb"] = float(tokens[mbpos[1]])
                    elif line.startswith("Pmy_natom="):
                        dims.update(my_natom=int(line.replace("Pmy_natom=", "").strip()))
                        #print("my_natom", dims["my_natom"])
                    else:
                        if line and line[0] == "-": line = line[1:]
                        tokens = grouper(2, line.replace("=", "").split())
                        if verbose > 1: print("tokens:", tokens)
                        dims.update([(t[0], int(t[1])) for t in tokens])

            return dims_dataset, spginfo_dataset

    def next_gs_scf_cycle(self) -> GroundStateScfCycle:
        """
        Return the next :class:`GroundStateScfCycle` in the file. None if not found.
        """
        return GroundStateScfCycle.from_stream(self)

    def get_all_gs_scf_cycles(self) -> list[GroundStateScfCycle]:
        """Return list of :class:`GroundStateScfCycle` objects. Empty list if no entry is found."""
        # NOTE: get_all should not used with next because of the call to self.seek(0)
        # The API should be refactored
        cycles = []
        self.seek(0)
        while True:
            cycle = self.next_gs_scf_cycle()
            if cycle is None: break
            cycles.append(cycle)

        self.seek(0)
        return cycles

    def next_d2de_scf_cycle(self) -> D2DEScfCycle:
        """
        Return :class:`D2DEScfCycle` with information on the DFPT iterations. None if not found.
        """
        return D2DEScfCycle.from_stream(self)

    def get_all_d2de_scf_cycles(self) -> list[D2DEScfCycle]:
        """Return list of :class:`D2DEScfCycle` objects. Empty list if no entry is found."""
        cycles = []
        self.seek(0)
        while True:
            cycle = self.next_d2de_scf_cycle()
            if cycle is None: break
            cycles.append(cycle)
        return cycles

    def plot(self, tight_layout=True, with_timer=False, show=True):
        """
        Plot GS/DFPT SCF cycles and timer data found in the output file.

        Args:
            with_timer: True if timer section should be plotted
        """
        from abipy.tools.plotting import MplExposer #, PanelExposer
        with MplExposer(slide_mode=False, slide_timeout=5.0) as e:
            e(self.yield_figs(tight_layout=tight_layout, with_timer=with_timer))

    # TODO: Use header and vars to understand if we have SCF/DFPT/Relaxation
    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        tight_layout = kwargs.pop("tight_layout", False)
        with_timer = kwargs.pop("with_timer", True)

        for icycle, cycle in enumerate(self.get_all_gs_scf_cycles()):
            yield cycle.plot(title=f"SCF cycle {icycle}", tight_layout=tight_layout, show=False)

        for icycle, cycle in enumerate(self.get_all_d2de_scf_cycles()):
            yield cycle.plot(title=f"DFPT cycle {icycle}", tight_layout=tight_layout, show=False)

        if with_timer:
            self.seek(0)
            try:
                yield self.get_timer().plot_all(tight_layout=tight_layout, show=False)
            except Exception:
                print("Abinit output files does not contain timopt data")

    def yield_plotly_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of plotly figures with minimal input from the user.
        """
        with_timer = kwargs.pop("with_timer", True)

        for icycle, cycle in enumerate(self.get_all_gs_scf_cycles()):
            yield cycle.plotly(title=f"SCF cycle {icycle}", show=False)

        for icycle, cycle in enumerate(self.get_all_d2de_scf_cycles()):
            yield cycle.plotly(title=f"DFPT cycle {icycle}", show=False)

        #if with_timer:
        #    self.seek(0)
        #    try:
        #        yield self.get_timer().plot_all(tight_layout=tight_layout, show=False)
        #    except Exception:
        #        print("Abinit output files does not contain timopt data")

    def compare_gs_scf_cycles(self, others, show=True) -> list[Figure]:
        """
        Produce and returns a list of matplotlib_ figure comparing the GS self-consistent
        cycle in self with the ones in others.

        Args:
            others: list of :class:`AbinitOutputFile` objects or strings with paths to output files.
            show: True to diplay plots.
        """
        # Open file here if we receive a string. Files will be closed before returning
        close_files = []
        for i, other in enumerate(others):
            if is_string(other):
                others[i] = self.__class__.from_file(other)
                close_files.append(i)

        fig, figures = None, []
        while True:
            cycle = self.next_gs_scf_cycle()
            if cycle is None: break

            fig = cycle.plot(show=False)
            for i, other in enumerate(others):
                other_cycle = other.next_gs_scf_cycle()
                if other_cycle is None: break
                last = (i == len(others) - 1)
                fig = other_cycle.plot(ax_list=fig.axes, show=show and last)
                if last:
                    fig.tight_layout()
                    figures.append(fig)

        self.seek(0)
        for other in others: other.seek(0)

        if close_files:
            for i in close_files: others[i].close()

        return figures

    def compare_d2de_scf_cycles(self, others, show=True) -> list[Figure]:
        """
        Produce and returns a list of matplotlib_ figure comparing the DFPT self-consistent
        cycle in self with the ones in others.

        Args:
            others: list of :class:`AbinitOutputFile` objects or strings with paths to output files.
            show: True to diplay plots.
        """
        # Open file here if we receive a string. Files will be closed before returning
        close_files = []
        for i, other in enumerate(others):
            if is_string(other):
                others[i] = self.__class__.from_file(other)
                close_files.append(i)

        fig, figures = None, []
        while True:
            cycle = self.next_d2de_scf_cycle()
            if cycle is None: break

            fig = cycle.plot(show=False)
            for i, other in enumerate(others):
                other_cycle = other.next_d2de_scf_cycle()
                if other_cycle is None: break
                last = (i == len(others) - 1)
                fig = other_cycle.plot(ax_list=fig.axes, show=show and last)
                if last:
                    fig.tight_layout()
                    figures.append(fig)

        self.seek(0)
        for other in others: other.seek(0)

        if close_files:
            for i in close_files: others[i].close()

        return figures

    def get_panel(self, **kwargs):
        """
        Build panel with widgets to interact with the Abinit output file either in a notebook or in panel app.
        """
        from abipy.panels.outputs import AbinitOutputFilePanel
        return AbinitOutputFilePanel(self).get_panel(**kwargs)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to nbpath. If ``nbpath`` is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("abo = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("print(abo.events)"),
            nbv.new_code_cell("abo.plot()"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


def validate_output_parser(abitests_dir=None, output_files=None) -> int:  # pragma: no cover
    """
    Validate/test Abinit output parser.

    Args:
        dirpath: Abinit tests directory.
        output_files: List of Abinit output files.

    Return: Exit code.
    """
    def is_abinit_output(path):
        """
        True if path is one of the output files used in the Abinit Test suite.
        """
        if not path.endswith(".abo"): return False
        if not path.endswith(".out"): return False

        with open(path, "rt") as fh:
            for i, line in enumerate(fh):
                if i == 1:
                    return line.rstrip().lower().endswith("abinit")
            return False

    # Files are collected in paths.
    paths = []

    if abitests_dir is not None:
        print("Analyzing directory %s for input files" % abitests_dir)

        for dirpath, dirnames, filenames in os.walk(abitests_dir):
            for fname in filenames:
                path = os.path.join(dirpath, fname)
                if is_abinit_output(path): paths.append(path)

    if output_files is not None:
        print("Analyzing files:", str(output_files))
        for arg in output_files:
            if is_abinit_output(arg): paths.append(arg)

    nfiles = len(paths)
    if nfiles == 0:
        cprint("Empty list of input files.", "red")
        return 0

    print("Found %d Abinit output files" % len(paths))
    errpaths = []
    for path in paths:
        print(path + ": ", end="")
        try:
            out = AbinitOutputFile.from_file(path)
            s = out.to_string(verbose=2)
            assert out.run_completed
            cprint("OK", "green")
        except Exception as exc:
            if not isinstance(exc, NotImplementedError):
                cprint("FAILED", "red")
                errpaths.append(path)
                import traceback
                print(traceback.format_exc())
                #print("[%s]: Exception:\n%s" % (path, str(exc)))
                #with open(path, "rt") as fh:
                #    print(10*"=" + "Input File" + 10*"=")
                #    print(fh.read())
                #    print()
            else:
                cprint("NOTIMPLEMENTED", "magenta")

    if errpaths:
        cprint("failed: %d/%d [%.1f%%]" % (len(errpaths), nfiles, 100 * len(errpaths)/nfiles), "red")
        for i, epath in enumerate(errpaths):
            cprint("[%d] %s" % (i, epath), "red")
    else:
        cprint("All input files successfully parsed!", "green")

    return len(errpaths)


class AboRobot(Robot):
    """
    This robot analyzes the results contained in multiple Abinit output files.
    Can compare dimensions, SCF cycles, analyze timers.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: AboRobot
    """
    EXT = "abo"

    def get_dims_dataframe(self, with_time=True, index=None) -> pd.DataFrame:
        """
        Build and return a |pandas-DataFrame| with the dimensions of the calculation.

        Args:
            with_time: True if walltime and cputime should be added
            index: Index of the dataframe. Use relative paths of files if None.
        """
        rows, my_index = [], []
        for i, abo in enumerate(self.abifiles):
            try:
                dims_dataset, spg_dataset = abo.get_dims_spginfo_dataset()
            except Exception as exc:
                cprint("Exception while trying to get dimensions from %s\n%s" % (abo.relpath, str(exc)), "yellow")
                continue

            for dtindex, dims in dims_dataset.items():
                dims = dims.copy()
                dims.update({"dtset": dtindex})
                # Add walltime and cputime in seconds
                if with_time:
                    dims.update(OrderedDict([(k, getattr(abo, k)) for k in
                        ("overall_cputime", "proc0_cputime", "overall_walltime", "proc0_walltime")]))
                rows.append(dims)
                my_index.append(abo.relpath if index is None else index[i])

        return pd.DataFrame(rows, index=my_index, columns=list(rows[0].keys()))

    def get_dataframe(self, with_geo=True, with_dims=True, abspath=False, funcs=None) -> pd.DataFrame:
        """
        Return a |pandas-DataFrame| with the most important results and the filenames as index.

        Args:
            with_geo: True if structure info should be added to the dataframe
            with_dims: True if dimensions should be added
            abspath: True if paths in index should be absolute. Default: Relative to getcwd().
            funcs: Function or list of functions to execute to add more data to the DataFrame.
                Each function receives a |GsrFile| object and returns a tuple (key, value)
                where key is a string with the name of column and value is the value to be inserted.
        """
        rows, row_names = [], []
        for label, abo in self.items():
            row_names.append(label)
            d = OrderedDict()

            if with_dims:
                dims_dataset, spg_dataset = abo.get_dims_spginfo_dataset()
                if len(dims_dataset) > 1:
                    cprint("Multiple datasets are not supported. ARGH!", "yellow")
                d.update(dims_dataset[1])

            # Add info on structure.
            if with_geo and abo.run_completed:
                d.update(abo.final_structure.get_dict4pandas(with_spglib=True))

            # Execute functions
            if funcs is not None: d.update(self._exec_funcs(funcs, abo))
            rows.append(d)

        row_names = row_names if not abspath else self._to_relpaths(row_names)
        return pd.DataFrame(rows, index=row_names, columns=list(rows[0].keys()))

    def get_time_dataframe(self) -> pd.DataFrame:
        """
        Return a |pandas-DataFrame| with the wall-time, cpu time in seconds and the filenames as index.
        """
        rows, row_names = [], []

        for label, abo in self.items():
            row_names.append(label)
            d = OrderedDict([(k, getattr(abo, k)) for k in
                ("overall_cputime", "proc0_cputime", "overall_walltime", "proc0_walltime")])
            rows.append(d)

        return pd.DataFrame(rows, index=row_names, columns=list(rows[0].keys()))

    # TODO
    #def gridplot_timer(self)

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield None

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to nbpath. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.AboRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
            nbv.new_code_cell("# robot.get_dims_dataframe()"),
            nbv.new_code_cell("robot.get_dataframe()"),
        ])

        # Mixins
        nb.cells.extend(self.get_baserobot_code_cells())

        return self._write_nb_nbpath(nb, nbpath)


class OutNcFile(AbinitNcFile):
    """
    Class representing the _OUT.nc file containing the dataset results
    produced at the end of the run. The netcdf variables can be accessed
    via instance attribute e.g. ``outfile.ecut``. Provides integration with ipython_.
    """
    # TODO: This object is deprecated

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = NetcdfReader(filepath)
        self._varscache = {k: None for k in self.reader.rootgrp.variables}

    def __dir__(self):
        """Ipython integration."""
        return sorted(list(self._varscache.keys()))

    def __getattribute__(self, name):
        try:
            return super().__getattribute__(name)
        except AttributeError:
            # Look in self._varscache
            varscache = super().__getattribute__("_varscache")
            if name not in varscache:
                raise AttributeError("Cannot find attribute %s" % name)
            reader = super().__getattribute__("reader")
            if varscache[name] is None:
                varscache[name] = reader.read_value(name)
            return varscache[name]

    @lazy_property
    def params(self) -> dict:
        """dict with parameters that might be subject to convergence studies."""
        return {}

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    def get_allvars(self):
        """
        Read all netcdf_ variables present in the file.
        Return dictionary varname --> value
        """
        for k, v in self._varscache.items():
            if v is not None: continue
            self._varscache[k] = self.reader.read_value(k)
        return self._varscache
